import pathlib
import pickle
from enum import Enum
from multiprocessing import Process

import fire
import matplotlib as mpl
from magicgui import magicgui

from cellfinder_visualize.process_summary import plot_cellfinder_bar_summary
from cellfinder_visualize.render import render_areas

mpl.use("Qt5Agg")


class Hemisphere(Enum):
    left = "left"
    right = "right"
    both = "both"


class ShaderStyle(Enum):
    cartoon = ("cartoon",)
    metallic = ("metallic",)
    plastic = ("plastic",)
    shiny = ("shiny",)
    glossy = "glossy"


@magicgui(
    experiment_dir={"mode": "d"},
    output_dir={"mode": "d"},
    config={"mode": "w"},
    call_button="Run",
    persist=False,
    tooltips=True,
    experiment_group={"choices": ["", ""], "allow_multiple": True},
    add_to_group_a=dict(widget_type="PushButton", text="Set Group A"),
    add_to_group_b=dict(widget_type="PushButton", text="Set Group B"),
    save_settings=dict(widget_type="PushButton", text="Save Settings"),
)
def analyse(
    add_to_group_a,
    add_to_group_b,
    save_settings,
    experiment_dir=pathlib.Path.home(),
    output_dir=pathlib.Path.home(),
    config=pathlib.Path.home(),
    coronal_slice_start=0,
    coronal_slice_end=12000,
    root=True,
    show_reference_structures=True,
    filter_cells_by_structure=False,
    hemisphere=Hemisphere.right,
    slice_root=True,
    subsample_factor=10,
    highlight_subregion="5",
    region_list=[
        "VISp",
        "VISpor",
        "VISli",
        "VISl",
        "VISal",
        "VISrl",
        "VISam",
        "VISpm",
        "CP",
    ],
    colors=[
        "#e41a1c",
        "#377eb8",
        "#4daf4a",
        "#984ea3",
        "#ff7f00",
        "#ffff33",
        "#a65628",
        "#f781bf",
        "#999999",
    ],
    reference_region="CTX",
    brainrender=True,
    plot_each_sample=True,
    plot_group_analysis=True,
    load_additional_obj_files=True,
    experiment_group=[],
    camera_pos=(-15854, -61680, 23155),
    camera_viewup=(1, 0, -1),
    camera_clipping_range=(57281, 96305),
    shader_style=ShaderStyle.cartoon,
):
    """
    :param experiment_dir: The directory containing all cellfinder output
    directories to analyse together.
    :param output_dir: The directory where all output figures will be saved.
    :param coronal_slice_start: To render a subvolume (e.g. to display a
    coronal slice) this indicates the beginning of the
    slice (0 is front of bulb)
    :param coronal_slice_end: To render a subvolume (e.g. to display a
    coronal slice) this indicates the end of the slice
    :param root: If true will render the whole brain in addition to the
    user-specified regions.
    :param show_reference_structures:  If True will render reference
    structures in addition to user-specified regions.
    :param filter_cells_by_structure: If True, will only display cells that
    fall within user-specified rendered regions
    :param hemisphere: Specify which hemisphere to render regions and cells
    in. Will also only compute cell counts for the selected hemisphere.
    :param slice_root: If True, does not show root brain in regions where no
    cells and subregions are displayed.
    :param subsample_factor: Show every nth cell
    :param highlight_subregion:
    :param region_list: The list of regions to render
    :param colors: The corresponding colors for these regions
    :param reference_region: The region used to normalise cell counts to.
    :param brainrender: If True, will generate a brainrender view.
    :param barplots: If True, will generate barplots for the cell counts.
    :param load_additional_obj_files: If True, any .obj files in the
    directory tree will be rendered.
    :return:
    """
    atlas_name = "allen_mouse_10um"
    camera = {
        "pos": camera_pos,
        "viewup": camera_viewup,
        "clippingRange": camera_clipping_range,
    }
    if brainrender:
        p = Process(
            target=render_areas,
            args=(
                analyse.group_a["points"],
                analyse.group_b["points"],
                region_list,
                colors,
                analyse.group_a["renderable objects"],
                filter_cells_by_structure,
                coronal_slice_start,
                coronal_slice_end,
                root,
                show_reference_structures,
                hemisphere.value,
                slice_root,
                highlight_subregion,
                subsample_factor,
                atlas_name,
                camera,
                shader_style.name,
            ),
        )
        p.start()

    plot_cellfinder_bar_summary(
        analyse.group_a["summary"],
        analyse.group_b["summary"],
        region_list,
        reference_region,
        output_dir,
        lateralisation=hemisphere,
        colors=colors,
        plot_each_sample=plot_each_sample,
        plot_group_analysis=plot_group_analysis,
        atlas_name="allen_mouse_10um",
    )


def is_cellfinder_path(p):
    return len(list(p.glob("cellfinder.json"))) > 0


@analyse.experiment_dir.changed.connect
def load_all_samples():
    p = analyse.experiment_dir.value
    paths = list(p.glob("*"))
    cellfinder_paths = [p for p in paths if is_cellfinder_path(p)]

    analyse.experiment_group.choices = cellfinder_paths


@analyse.save_settings.changed.connect
def save_settings(event=None):
    print(analyse.asdict())

    with open(
        pathlib.Path(analyse.output_dir.value) / "settings.pkl", "wb"
    ) as f:
        pickle.dump(analyse.asdict(), f)


@analyse.config.changed.connect
def load_settings(event=None):
    with open(analyse.config.value, "rb") as f:
        loaded_dict = pickle.load(f)

    for k, v in loaded_dict.items():
        analyse[k].value = v


@analyse.add_to_group_a.changed.connect
def add_to_group_a(event=None):
    group_dict = get_file_paths_for_group(analyse)
    analyse.group_a = group_dict
    print(analyse.group_a)


@analyse.add_to_group_b.changed.connect
def add_to_group_b(event=None):
    group_dict = get_file_paths_for_group(analyse)
    analyse.group_b = group_dict
    print(analyse.group_b)


def get_file_paths_for_group(widget):
    all_summary_files = []
    all_points_files = []
    render_files = []
    for p in widget.experiment_group.value:
        all_summary_files.extend(p.rglob("*summary*.csv"))
        all_points_files.extend(p.rglob("*points*npy"))
        render_files.extend(p.rglob("*.obj"))
    group_dict = {}
    group_dict.setdefault("summary", all_summary_files)
    group_dict.setdefault("points", all_points_files)
    group_dict.setdefault("renderable objects", render_files)
    return group_dict


def main():
    fire.Fire(analyse.show(run=True))


if __name__ == "__main__":
    main()



import logging
import pathlib

import bg_atlasapi
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from makefig.config import default_axis_positions, default_label_positions
from makefig.construct_figure import make_figure

from cellfinder_visualize.region_groupings import metrics_and_axis_labels


def get_n_cells_in_region(
    atlas, df_results, key, distance=2, lateralisation="left"
):
    structure_acronyms = atlas.get_structure_descendants(key)
    structure_names = [atlas.structures[k]["name"] for k in structure_acronyms]
    total = 0
    total_volume = 0

    sum_key, volume_key = get_lateralisation_keys(lateralisation)

    for structure_name in list(structure_names):
        child_result_row = df_results[
            df_results["structure_name"] == structure_name
        ]
        if len(child_result_row) > 0:
            count = child_result_row[sum_key].values[0]
            volume = child_result_row[volume_key].values[0]
            total += count
            total_volume += volume
            logging.info(
                f"counted {count} cells in {structure_name}, {sum_key} in "
                f"{key} found to be {total}.. "
                f"total per mm {total/total_volume}"
            )
    return total, total_volume


def get_lateralisation_keys(lateralisation):
    if lateralisation == "left":
        sum_key = "left_cell_count"
        volume_key = "left_volume_mm3"
    elif lateralisation == "right":
        sum_key = "right_cell_count"
        volume_key = "right_volume_mm3"
    else:
        sum_key = "total_cells"
        volume_key = "total_volume_mm3"
    return sum_key, volume_key


def get_cellfinder_bar_data(
    atlas,
    experiment_filepath,
    plotting_keys,
    reference_key,
    sample_id,
    lateralisation="left",
):
    sum_key, volume_key = get_lateralisation_keys(lateralisation)

    df_results = pd.read_csv(experiment_filepath)
    cells_sum = df_results[sum_key].sum()

    df_dict = {}
    totals = []
    region_volumes = []
    reference_counts = []
    reference_keys = []

    for k, ref_k in zip(plotting_keys, [reference_key] * len(plotting_keys)):
        n_cells_in_region, region_volume = get_n_cells_in_region(
            atlas, df_results, k, lateralisation=lateralisation
        )
        n_cells_in_reference, _ = get_n_cells_in_region(
            atlas, df_results, ref_k, distance=0, lateralisation=lateralisation
        )
        reference_counts.append(n_cells_in_reference)
        reference_keys.append(ref_k)
        region_volumes.append(region_volume)
        totals.append(n_cells_in_region)

    df_dict.setdefault("n_cells_in_region", totals)
    df_dict.setdefault("n_cells_in_reference_region", reference_counts)
    df_dict.setdefault("reference_regions", reference_keys)
    df_dict.setdefault(
        "percent_of_reference_region",
        np.array(totals) / np.array(reference_counts) * 100,
    )
    df_dict.setdefault(
        f"n_cells_in_whole_brain_{sum_key}",
        [cells_sum] * len(reference_counts),
    )
    df_dict.setdefault(
        "cells_per_mm3", np.array(totals) / np.array(region_volumes)
    )
    df_dict.setdefault("percentage", np.array(totals) / cells_sum * 100)
    df_dict.setdefault("region", plotting_keys)
    df_dict.setdefault("sample_id", [sample_id] * len(reference_counts))

    single_sample_df = pd.DataFrame.from_dict(df_dict)

    single_sample_df["percent_reference_labels"] = (
        single_sample_df["region"]
        + " / "
        + single_sample_df["reference_regions"]
    )

    return single_sample_df


def adjust_bar_width(ax, new_value):
    """https://stackoverflow.com/questions/34888058/
    changing-width-of-bars-in-bar-chart-created-using-seaborn-factorplot"""

    for patch in ax.patches:
        current_width = patch.get_width()
        diff = current_width - new_value

        patch.set_width(new_value)
        patch.set_x(patch.get_x() + diff * 0.5)


def plot_pooled_experiments(
    df_group_a,
    df_group_b,
    reference_structure_key,
    output_directory,
    boxplot=False,
):
    if len(df_group_a) > 1 and len(df_group_b) > 1:
        df_group_a = pd.concat(df_group_a)
        df_group_b = pd.concat(df_group_b)

        if boxplot:
            h_fig, axes_dict = make_figure(
                default_label_positions,
                default_axis_positions,
                axes=("A", "B", "C", "D"),
            )
            for metric, ax in zip(
                metrics_and_axis_labels.items(), axes_dict.values()
            ):
                plt.sca(ax)
                plot_boxplots(
                    df_group_a, df_group_b, metric=metric[0], label=metric[1]
                )
        else:
            for all_samples_df in [df_group_a, df_group_b]:
                h_fig, axes_dict = make_figure(
                    default_label_positions,
                    default_axis_positions,
                    axes=("A", "B", "C", "D"),
                )
                for metric, ax in zip(
                    metrics_and_axis_labels.items(), axes_dict.values()
                ):
                    plt.sca(ax)
                    average_counts_df = (
                        all_samples_df.groupby("region")
                        .agg(avg=(metric[0], "mean"))
                        .reset_index()
                    )
                    region_labels = all_samples_df["region"].unique()

                    for i, region_label in enumerate(region_labels):
                        values = all_samples_df.query(
                            f'region == "{region_label}"'
                        )[metric[0]]
                        avg = average_counts_df.query(
                            f'region == "{region_label}"'
                        )["avg"].values[0]
                        plt.plot([i] * len(values), values, "o", alpha=0.5)
                        plt.hlines(avg, i - 0.2, i + 0.2, color="k")
                    plt.xlim([-1, len(region_labels)])

                    plt.ylabel(metric[1])

                    if metric[0] == "percent_of_reference_region":
                        labels = [
                            label + " / " + reference_structure_key
                            for label in region_labels
                        ]
                        plt.xticks(
                            range(len(region_labels)),
                            labels=labels,
                            rotation=45,
                        )
                        plt.xlabel("Region / Reference Region")

                    else:
                        plt.xticks(
                            range(len(region_labels)),
                            labels=region_labels,
                            rotation=45,
                        )
                        plt.xlabel("Region")
                if output_directory is not None:
                    save_output(
                        h_fig,
                        output_directory,
                        reference_structure_key,
                        all_samples_df,
                        fig_type="all_samples",
                    )
                plt.show()


def plot_boxplots(df_group_a, df_group_b, metric, label):
    names = []
    all_values_a = []
    all_values_b = []
    bar_space = 0.5
    padding = 0.1

    for i, (name, group) in enumerate(
        df_group_a[["region", metric]].groupby("region")
    ):
        values = group[metric].values
        names.append(name)
        all_values_a.append(values)

    bp = plt.boxplot(
        all_values_a, positions=range(0, len(all_values_a) * 2, 2)
    )
    for item in ["boxes", "whiskers", "fliers", "caps"]:
        plt.setp(bp[item], color="k")

    for i, (name, group) in enumerate(
        df_group_b[["region", metric]].groupby("region")
    ):
        values = group[metric].values
        all_values_b.append(values)

    bp = plt.boxplot(
        all_values_b,
        positions=np.arange(0, len(all_values_b) * 2, 2) + bar_space + padding,
    )
    for item in ["boxes", "whiskers", "fliers", "caps"]:
        plt.setp(bp[item], color="r")
    plt.ylabel(label)

    plt.xticks(
        np.arange(0, len(all_values_a) * 2, 2) + bar_space / 2,
        names,
        rotation=45,
    )

    plt.show()


def plot_cellfinder_bar_summary(
    group_a_filepaths,
    group_b_filepaths,
    plotting_keys,
    reference_structure_key,
    output_directory,
    lateralisation,
    colors,
    print_latex=False,
    plot_each_sample=False,
    plot_group_analysis=False,
    atlas_name="allen_mouse_10um",
):
    atlas = bg_atlasapi.BrainGlobeAtlas(atlas_name)
    dfs_all = []
    colors_palette = sns.set_palette(sns.color_palette(colors))
    for experiment_filepaths in [group_a_filepaths, group_b_filepaths]:
        group_dfs = []

        for experiment_filepath in experiment_filepaths:
            single_sample_df = get_cellfinder_bar_data(
                atlas,
                experiment_filepath,
                plotting_keys,
                reference_structure_key,
                pathlib.Path(experiment_filepath).stem,
                lateralisation=lateralisation,
            )

            group_dfs.append(single_sample_df)

            if plot_each_sample:
                h_fig, axes_dict = make_figure(
                    default_label_positions,
                    default_axis_positions,
                    axes=("A", "B", "C", "D"),
                )
                plot_single_sample(
                    single_sample_df,
                    axes_dict,
                    colors_palette,
                    experiment_filepath,
                    lateralisation,
                    plotting_keys,
                    reference_structure_key,
                )

                if output_directory is not None:
                    save_output(
                        h_fig,
                        output_directory,
                        reference_structure_key,
                        single_sample_df,
                        fig_type=f"{experiment_filepath.parent.stem}",
                    )

            if print_latex:
                print_latex_table(single_sample_df)

        dfs_all.append(group_dfs)

    if plot_group_analysis:
        plot_pooled_experiments(
            dfs_all[0],
            dfs_all[1],
            reference_structure_key,
            output_directory,
            boxplot=True,
        )
        plt.pause(0.0001)
        plot_pooled_experiments(
            dfs_all[0],
            dfs_all[1],
            reference_structure_key,
            output_directory,
            boxplot=False,
        )


def plot_single_sample(
    single_sample_df,
    axes_dict,
    colors_palette,
    experiment_filepath,
    lateralisation,
    plotting_keys,
    reference_structure_key,
):
    plt.suptitle(
        f"Sample: {pathlib.Path(experiment_filepath).parent.parent.stem}"
    )

    for metric, ax in zip(metrics_and_axis_labels.items(), axes_dict.values()):
        plt.sca(ax)

        if metric[0] == "percent_of_reference_region":
            sns.barplot(
                data=single_sample_df,
                x="percent_reference_labels",
                y=metric[0],
                palette=colors_palette,
            )
            plt.xlabel("Region / Reference Region")

        else:
            sns.barplot(
                data=single_sample_df,
                x="region",
                y=metric[0],
                palette=colors_palette,
            )
            plt.xlabel("Region")

        plt.xlim([-1, len(plotting_keys)])
        plt.xticks(rotation=45)
        plt.ylabel(metric[1])

        plt.ion()
        plt.show()
        plt.pause(0.0001)


def print_latex_table(single_sample_df):
    single_sample_df["percentage"] = single_sample_df["percentage"]
    single_sample_df["percentage"] = single_sample_df["percentage"].round(2)
    single_sample_df["cells_per_mm3"] = single_sample_df[
        "cells_per_mm3"
    ].round(1)
    single_sample_df["n_cells_in_region"] = single_sample_df[
        "n_cells_in_region"
    ].astype(int)
    df2 = single_sample_df[
        ["region", "n_cells_in_region", "percentage", "cells_per_mm3"]
    ]
    print(df2.to_latex())


def save_output(
    fig,
    output_directory,
    reference_structure_key,
    single_brain_df,
    fig_type="",
):
    output_directory = pathlib.Path(output_directory)
    fig.savefig(output_directory / f"{reference_structure_key}_{fig_type}.png")
    single_brain_df.to_csv(
        output_directory / f"{reference_structure_key}_{fig_type}.csv"
    )



import numpy as np
from brainrender.actors import Points


def render_cells_in_regions(cells, regions, regions_rendered, scene, color):
    for region in regions:
        if region is not None:
            regions_rendered = render_cells_in_region(
                cells, region, regions_rendered, scene, color
            )
    return regions_rendered


def render_regions(_colors, region_keys, scene, hemisphere, alpha=0.2):
    regions = []

    for region_name, color in zip(region_keys, _colors):
        region = scene.add_brain_region(
            region_name, color=color, alpha=alpha, hemisphere=hemisphere
        )
        regions.append(region)
    return regions


def highlight_layer(
    atlas,
    highlight_substructure_key,
    region_name,
    regions_rendered,
    scene,
    hemisphere,
):
    children = atlas.get_structure_descendants(region_name)

    if highlight_substructure_key != "all":
        for k in children:
            if k != region_name:
                if str(highlight_substructure_key) in k:
                    add_substructure_region(
                        hemisphere, k, regions_rendered, scene
                    )
    else:
        for k in children:
            if k != region_name:
                add_substructure_region(hemisphere, k, regions_rendered, scene)


def add_substructure_region(hemisphere, k, regions_rendered, scene):
    r = scene.add_brain_region(k, color="w", alpha=0.3, hemisphere=hemisphere)
    scene.add_silhouette(r, color="r", lw=2)
    regions_rendered.append(r)


def render_cells_in_region(cells, region, regions_rendered, scene, color):
    cells_in_region = region.mesh.insidePoints(cells).points()
    cells_in_region = Points(
        cells_in_region, radius=20, colors=color, alpha=0.5
    )
    scene.add(cells_in_region)
    regions_rendered.append(cells_in_region)
    return regions_rendered


def remove_unwanted_hemisphere(lateralisation, regions_rendered, scene):
    direction = -1 if lateralisation == "left" else 1
    plane = scene.atlas.get_plane(norm=(0, 0, direction))
    regions_rendered = [x for x in regions_rendered if x is not None]
    scene.slice(plane, actors=regions_rendered, close_actors=True)


def slice_coronal_volume(
    coronal_slice, regions_rendered, scene, slice_thickness
):
    xyz = np.array(
        [0.00000000, 3829.52651499, 5682.68089654]
    )  # abitrary point
    xyz[0] += coronal_slice

    plane = scene.atlas.get_plane(pos=xyz, norm=(1, 0, 0))
    regions_rendered = [x for x in regions_rendered if x is not None]
    scene.slice(plane, actors=regions_rendered, close_actors=True)
    if slice_thickness is not None:
        xyz[0] += slice_thickness
        plane2 = scene.atlas.get_plane(pos=xyz, norm=(-1, 0, 0))
        scene.slice(plane2, actors=regions_rendered, close_actors=True)



reference_structures_to_render = ["TH", "SCs", "SCm", "PAG"]

metrics_and_axis_labels = {
    "n_cells_in_region": "Cells ( Number )",
    "percentage": "Cells ( % )",
    "cells_per_mm3": "Cell density ( cells / mm3 )",
    "percent_of_reference_region": "Cells ( % of ref )",
}

camera = {
    "pos": (-15854, -61680, 23155),
    "viewup": (1, 0, -1),
    "clippingRange": (57281, 96305),
}

zoom = 1.5

additional_obj_color = "k"



import bg_atlasapi
import brainrender
import numpy as np
from brainrender import Scene

from cellfinder_visualize.region_groupings import (
    additional_obj_color,
    camera,
    reference_structures_to_render,
    zoom,
)
from cellfinder_visualize.rendering_functions import (
    highlight_layer,
    remove_unwanted_hemisphere,
    render_cells_in_region,
    render_cells_in_regions,
    render_regions,
    slice_coronal_volume,
)


def render_areas(
    points_files_a,
    points_files_b,
    region_keys,
    colors,
    additional_obj_files=None,
    filter_cells_by_structure=False,
    coronal_slice=None,
    slice_thickness=None,
    root=True,
    show_reference_structures=None,
    hemisphere="right",
    slice_root=True,
    highlight_subregion="6",
    subsample_factor=10,
    atlas_name="allen_mouse_10um",
    camera=camera,
    shader_style="cartoon",
):
    atlas = bg_atlasapi.BrainGlobeAtlas(atlas_name)

    brainrender.settings.SHADER_STYLE = shader_style

    regions_rendered = []
    scene = Scene(title="labelled cells", root=root)

    cells_group_a = load_cells_in_group(points_files_a, subsample_factor)
    cells_group_b = load_cells_in_group(points_files_b, subsample_factor)

    if slice_root:
        regions_rendered.append(scene.root)

    for region_name in region_keys:
        highlight_layer(
            atlas,
            highlight_subregion,
            region_name,
            regions_rendered,
            scene,
            hemisphere,
        )

    if not filter_cells_by_structure:
        render_cells_in_region(
            cells_group_a, scene.root, regions_rendered, scene, color="k"
        )
        render_cells_in_region(
            cells_group_b, scene.root, regions_rendered, scene, color="r"
        )
        regions = render_regions(
            colors, region_keys, scene, hemisphere=hemisphere
        )
        regions_rendered.extend(regions)

    else:
        if show_reference_structures:
            for k in reference_structures_to_render:
                reg = scene.add_brain_region(
                    k, color="grey", alpha=0.3, hemisphere=hemisphere
                )
                regions_rendered.append(reg)

        regions = render_regions(colors, region_keys, scene, hemisphere)
        regions_rendered.extend(regions)

        regions_rendered = render_cells_in_regions(
            cells_group_a, regions, regions_rendered, scene, color="k"
        )
        regions_rendered = render_cells_in_regions(
            cells_group_b, regions, regions_rendered, scene, color="r"
        )

    if additional_obj_files is not None:
        for fpath in additional_obj_files:
            color = "b" if "fiber" in str(fpath) else additional_obj_color
            r = scene.add(fpath, color=color)
            regions_rendered.append(r)

    if hemisphere != "both":
        remove_unwanted_hemisphere(hemisphere, regions_rendered, scene)

    if coronal_slice is not None:
        regions_rendered.append(scene.root)
        slice_coronal_volume(
            coronal_slice, regions_rendered, scene, slice_thickness
        )
    scene.render(camera=camera, zoom=zoom)


def load_cells_in_group(points_files, subsample_factor):
    cells = []
    for points_file in points_files:
        these_cells = np.load(points_file)[::subsample_factor]
        cells.extend(these_cells)
    return cells
